""" Ability to push to a standard Git system without requiring Github or Gitlab.
    No pull requests will be automatically created. """

import argparse
import tsrc.git
from tsrc.cli.push import RepositoryInfo


class PushAction(tsrc.cli.push.PushAction):
    def __init__(
        self, repository_info: RepositoryInfo, args: argparse.Namespace
    ) -> None:
        super().__init__(repository_info, args)

    def setup_service(self) -> None:
        pass

    def post_push(self) -> None:
        pass



import argparse
from typing import Any, List

from gitlab import Gitlab
from path import Path
import mock
import pytest

import tsrc
from tsrc.test.helpers.cli import CLI
from tsrc.cli.push import RepositoryInfo
from tsrc.cli.push_gitlab import PushAction, FeatureNotAvailable


GITLAB_URL = "http://gitlab.example.com"

# Can't use 'name' in the ctor because it sets the name
# of the mock, not the 'name' attribute of the returned
# instance.
ALICE = mock.Mock(username="alice", id=1)
ALICE.name = "Alice Liddell"
BOB = mock.Mock(username="bob", id=2, name="")
BOB.name = "Bob SquarePants"
EVE = mock.Mock(username="eve", id=3)
EVE.name = "Eve Arden"


class UserList:
    def __init__(self, user_list: List[Any]) -> None:
        self.user_list = user_list
        self.index = -1
        self.total = len(user_list)
        self.next_page = None

    def __iter__(self) -> Any:
        return (user for user in self.user_list)


def execute_query(*, query: str, **kwargs: Any) -> UserList:
    matches = list()
    for user in [ALICE, BOB, EVE]:
        if user.username == query:
            matches.append(user)
    return UserList(matches)


def gitlab_mock_with_merge_requests(mock_merge_requests: List[Any]) -> Any:
    gitlab_mock = mock.Mock()

    mock_project = mock.Mock()
    mock_project.members.list.side_effect = execute_query
    mock_project.mergerequests.list.return_value = mock_merge_requests

    mock_group = mock.Mock()
    mock_group.members.list.side_effect = execute_query

    gitlab_mock.projects.get.return_value = mock_project

    gitlab_mock.groups.get.return_value = mock_group

    mock_feature = mock.Mock()
    mock_feature.name = "multiple_merge_request_assignees"

    gitlab_mock.features.list.return_value = [mock_feature]

    return gitlab_mock


def execute_push(
    repo_path: Path, push_args: argparse.Namespace, gitlab_mock: Gitlab
) -> None:
    workspace_mock = mock.Mock()
    workspace_mock.get_github_enterprise_url.return_value = None
    workspace_mock.get_gitlab_url.return_value = GITLAB_URL

    repository_info = RepositoryInfo.read(repo_path, workspace=workspace_mock)
    push_action = PushAction(repository_info, push_args, gitlab_api=gitlab_mock)
    push_action.execute()


def test_check_for_multiple_assignees_feature(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    gitlab_mock = gitlab_mock_with_merge_requests([])
    gitlab_mock.features.list.return_value = []
    push_args.reviewers = ["alice", "bob"]

    with pytest.raises(FeatureNotAvailable) as e:
        execute_push(repo_path, push_args, gitlab_mock)
    assert e.value.name == "multiple_merge_request_assignees"


def test_creating_merge_request_explicit_target_branch_with_assignee(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    tsrc.git.run(repo_path, "checkout", "-b", "new-feature")
    tsrc.git.run(repo_path, "commit", "--message", "new feature", "--allow-empty")

    gitlab_mock = gitlab_mock_with_merge_requests([])

    push_args.assignee = "alice"
    push_args.target_branch = "next"
    push_args.title = "Best feature ever"

    mock_project = gitlab_mock.projects.get()
    new_mr = mock.Mock()
    mock_project.mergerequests.create.return_value = new_mr
    execute_push(repo_path, push_args, gitlab_mock)

    mock_project.mergerequests.create.assert_called_once_with(
        {
            "source_branch": "new-feature",
            "target_branch": "next",
            "title": "new-feature",
        }
    )

    new_mr.title = "Best feature ever"
    new_mr.save.assert_called_once_with()


def test_creating_merge_request_uses_default_branch(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    gitlab_mock = gitlab_mock_with_merge_requests([])
    mock_project = gitlab_mock.projects.get()
    mock_project.default_branch = "devel"
    tsrc.git.run(repo_path, "checkout", "-b", "new-feature")
    tsrc.git.run(repo_path, "commit", "--message", "new feature", "--allow-empty")

    mock_project = gitlab_mock.projects.get()
    new_mr = mock.Mock(iid="43", web_url="http://43")
    mock_project.mergerequests.create.return_value = new_mr

    push_args.assignee = "alice"
    execute_push(repo_path, push_args, gitlab_mock)

    mock_project.mergerequests.create.assert_called_once_with(
        {
            "source_branch": "new-feature",
            "target_branch": "devel",
            "title": "new-feature",
        }
    )


def test_set_approvers(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:

    tsrc.git.run(repo_path, "checkout", "-b", "new-feature")
    tsrc.git.run(repo_path, "commit", "--message", "new feature", "--allow-empty")

    push_args.reviewers = ["alice", "eve"]

    mock_mr = mock.Mock(iid="42", web_url="http://42")
    gitlab_mock = gitlab_mock_with_merge_requests([mock_mr])

    execute_push(repo_path, push_args, gitlab_mock)

    mock_mr.approvals.set_approvers.assert_called_once_with([ALICE.id, EVE.id])
    mock_mr.save.assert_called_once()


def test_update_existing_merge_request(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    tsrc.git.run(repo_path, "checkout", "-b", "new-feature")
    tsrc.git.run(repo_path, "commit", "--message", "new feature", "--allow-empty")

    mock_mr = mock.Mock(iid="42", web_url="http://42", title="old title")
    gitlab_mock = gitlab_mock_with_merge_requests([mock_mr])

    push_args.target_branch = "next"
    push_args.title = "Best feature ever"
    execute_push(repo_path, push_args, gitlab_mock)

    gitlab_mock.mergerequests.create.assert_not_called()
    assert mock_mr.remove_source_branch
    assert mock_mr.target_branch == "next"
    assert mock_mr.title == "Best feature ever"


def test_close_merge_request(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    tsrc.git.run(repo_path, "checkout", "-b", "new-feature")
    tsrc.git.run(repo_path, "commit", "--message", "new feature", "--allow-empty")

    mock_mr = mock.Mock(iid="42", web_url="http://42", title="old title")
    gitlab_mock = gitlab_mock_with_merge_requests([mock_mr])

    push_args.close = True
    execute_push(repo_path, push_args, gitlab_mock)

    assert mock_mr.state_event == "close"
    mock_mr.save.assert_called_once()


def test_do_not_change_mr_target(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    mock_mr = mock.Mock(iid="42", web_url="http://42", target_branch="old-branch")
    gitlab_mock = gitlab_mock_with_merge_requests([mock_mr])

    execute_push(repo_path, push_args, gitlab_mock)

    assert mock_mr.target_branch == "old-branch"


def test_accept_merge_request(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    tsrc.git.run(repo_path, "checkout", "-b", "new-feature")
    tsrc.git.run(repo_path, "commit", "--message", "new feature", "--allow-empty")

    mock_mr = mock.Mock(iid="42", web_url="http://42", target_branch="old-branch")
    gitlab_mock = gitlab_mock_with_merge_requests([mock_mr])

    push_args.accept = True
    execute_push(repo_path, push_args, gitlab_mock)

    mock_mr.merge.assert_called_once_with(merge_when_pipeline_succeeds=True)


def test_unwipify_existing_merge_request(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    mock_mr = mock.Mock(
        title="WIP: nice title", web_url="http://example.com/42", iid=42
    )
    gitlab_mock = gitlab_mock_with_merge_requests([mock_mr])

    push_args.ready = True
    execute_push(repo_path, push_args, gitlab_mock)

    assert mock_mr.title == "nice title"
    mock_mr.save.assert_called_once()


def test_wipify_existing_merge_request(
    repo_path: Path, tsrc_cli: CLI, push_args: argparse.Namespace
) -> None:
    mock_mr = mock.Mock(title="not ready", web_url="http://example.com/42", iid=42)
    gitlab_mock = gitlab_mock_with_merge_requests([mock_mr])

    push_args.wip = True
    execute_push(repo_path, push_args, gitlab_mock)

    assert mock_mr.title == "WIP: not ready"
    mock_mr.save.assert_called_once()



import tsrc.cli.push_github
import argparse
from typing import Optional

from github3 import GitHub

import tsrc
import tsrc.github
from tsrc.cli.push import RepositoryInfo


class PushAction(tsrc.cli.push_github.PushAction):
    def __init__(
        self,
        repository_info: RepositoryInfo,
        args: argparse.Namespace,
        github_api: Optional[GitHub] = None,
    ) -> None:
        if not github_api:
            github_api = tsrc.github.login(
                github_enterprise_url=repository_info.repository_login_url
            )

        super().__init__(repository_info, args, github_api)



import argparse
import mock

from path import Path
import tsrc
from tsrc.cli.push_git import PushAction
from tsrc.cli.push import RepositoryInfo


def test_push_use_tracked_branch(
    repo_path: Path, push_args: argparse.Namespace
) -> None:
    tsrc.git.run(repo_path, "checkout", "-b", "local")
    tsrc.git.run(repo_path, "push", "-u", "origin", "local:remote")

    repository_info = RepositoryInfo.read(
        repo_path, workspace=mock_workspace_git_urls()
    )
    dummy_push = PushAction(repository_info, push_args)
    dummy_push.push()
    _, out = tsrc.git.run_captured(repo_path, "ls-remote")
    assert "local" not in out
    assert "heads/remote" in out


def test_push_use_given_push_spec(
    repo_path: Path, push_args: argparse.Namespace
) -> None:
    tsrc.git.run(repo_path, "checkout", "-b", "local")
    push_args.push_spec = "local:remote"
    repository_info = RepositoryInfo.read(
        repo_path, workspace=mock_workspace_git_urls()
    )
    dummy_push = PushAction(repository_info, push_args)
    dummy_push.push()
    _, out = tsrc.git.run_captured(repo_path, "ls-remote")
    assert "local" not in out
    assert "heads/remote" in out


def test_service_from_url() -> None:
    workspace_mock = mock_workspace_git_urls()

    def get_service(url: str) -> str:
        return tsrc.cli.push.service_from_url(url, workspace_mock)

    assert get_service("git@github.com:TankerHQ/tsrc.git") == "github"
    assert get_service("git@gitlab.ex.co:TankerHQ/tsrc.git") == "gitlab"
    assert get_service("git@github.ex.co:TankerHQ/tsrc.git") == "github_enterprise"
    assert get_service("git@git.ex.co:TankerHQ/tsrc.git") == "git"

    workspace_mock.get_github_enterprise_url.return_value = (
        "https://github.ex.co:8443/github"
    )
    workspace_mock.get_gitlab_url.return_value = "https://gitlab.ex.co:8443/gitlab"
    assert get_service("git@gitlab.ex.co:TankerHQ/tsrc.git") == "gitlab"
    assert get_service("git@github.ex.co:TankerHQ/tsrc.git") == "github_enterprise"
    assert get_service("git@gitlab.ex.co:8443:TankerHQ/tsrc.git") == "gitlab"
    assert get_service("git@github.ex.co:8443:TankerHQ/tsrc.git") == "github_enterprise"


def test_project_name_from_url() -> None:
    def project_name(url: str) -> str:
        return tsrc.cli.push.project_name_from_url(url)

    assert project_name("git@ex.co:foo/bar.git") == "foo/bar"
    assert project_name("ssh://git@ex.co:8022/foo/bar.git") == "foo/bar"


def mock_workspace_git_urls() -> mock.Mock:
    workspace_mock = mock.Mock(tsrc.Workspace)
    workspace_mock.get_github_enterprise_url.return_value = "https://github.ex.co/"
    workspace_mock.get_gitlab_url.return_value = "https://gitlab.ex.co/"

    return workspace_mock



""" Entry point for tsrc status """

from typing import Dict, List, Union

import argparse
import collections
import shutil

import cli_ui as ui
from path import Path

import tsrc
import tsrc.errors
import tsrc.cli
import tsrc.git


StatusOrError = Union[tsrc.git.Status, Exception]
CollectedStatuses = Dict[str, StatusOrError]


def describe_status(status: StatusOrError) -> List[ui.Token]:
    """ Returns a list of tokens suitable for ui.info() """
    if isinstance(status, tsrc.errors.MissingRepo):
        return [ui.red, "error: missing repo"]
    if isinstance(status, Exception):
        return [ui.red, "error: ", status]
    return describe_git_status(status)


def describe_git_status(git_status: tsrc.git.Status) -> List[ui.Token]:
    res = []  # type: List[ui.Token]
    res += describe_branch(git_status)
    res += describe_position(git_status)
    res += describe_dirty(git_status)
    return res


def describe_branch(git_status: tsrc.git.Status) -> List[ui.Token]:
    res = list()  # type: List[ui.Token]
    if git_status.branch:
        res += [ui.green, git_status.branch]
    elif git_status.sha1:
        res += [ui.red, git_status.sha1]
    if git_status.tag:
        res += [ui.reset, ui.brown, "on", git_status.tag]
    return res


def commit_string(number: int) -> str:
    if number == 1:
        return "commit"
    else:
        return "commits"


def describe_position(git_status: tsrc.git.Status) -> List[ui.Token]:
    res = []  # type: List[ui.Token]
    if git_status.ahead != 0:
        up = ui.Symbol("↑", "+")
        n_commits = commit_string(git_status.ahead)
        ahead_desc = "{}{} {}".format(up.as_string, git_status.ahead, n_commits)
        res += [ui.blue, ahead_desc, ui.reset]
    if git_status.behind != 0:
        down = ui.Symbol("↓", "-")
        n_commits = commit_string(git_status.behind)
        behind_desc = "{}{} {}".format(down.as_string, git_status.behind, n_commits)
        res += [ui.blue, behind_desc, ui.reset]
    return res


def describe_dirty(git_status: tsrc.git.Status) -> List[ui.Token]:
    res = []  # type: List[ui.Token]
    if git_status.dirty:
        res += [ui.red, "(dirty)"]
    return res


def erase_last_line() -> None:
    terminal_size = shutil.get_terminal_size()
    ui.info(" " * terminal_size.columns, end="\r")


class StatusCollector(tsrc.Task[tsrc.Repo]):
    def __init__(self, workspace_path: Path) -> None:
        self.workspace_path = workspace_path
        self.statuses = collections.OrderedDict()  # type: CollectedStatuses
        self.num_repos = 0

    def display_item(self, repo: tsrc.Repo) -> str:
        return repo.src

    def process(self, index: int, total: int, repo: tsrc.Repo) -> None:
        ui.info_count(index, total, repo.src, end="\r")
        full_path = self.workspace_path / repo.src

        if not full_path.exists():
            self.statuses[repo.src] = tsrc.errors.MissingRepo(repo.src)
            return

        try:
            self.statuses[repo.src] = tsrc.git.get_status(full_path)
        except Exception as e:
            self.statuses[repo.src] = e
        erase_last_line()

    def on_start(self, num_items: int) -> None:
        ui.info_1("Collecting statuses of %d repos" % num_items)
        self.num_repos = num_items

    def on_success(self) -> None:
        erase_last_line()
        if not self.statuses:
            ui.info_2("Workspace is empty")
            return
        ui.info_2("Workspace status:")
        max_src = max(len(x) for x in self.statuses.keys())
        for src, status in self.statuses.items():
            message = [ui.green, "*", ui.reset, src.ljust(max_src)]
            message += describe_status(status)
            ui.info(*message)


def main(args: argparse.Namespace) -> None:
    workspace = tsrc.cli.get_workspace(args)
    status_collector = StatusCollector(workspace.root_path)
    workspace.load_manifest()
    tsrc.run_sequence(workspace.get_repos(), status_collector)
